unit uMain;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, ExtDlgs;

type
  TfrmMain = class(TForm)
    imgPreview: TImage;
    btnOpen: TButton;
    memoExif: TMemo;
    lblExif: TLabel;
    btnExit: TButton;
    dlgOpen: TOpenPictureDialog;
    procedure btnOpenClick(Sender: TObject);
    procedure btnExitClick(Sender: TObject);
  private
    procedure ReadExif(const vFileName: String);
  public
  end;

var
  frmMain: TfrmMain;

implementation

uses
  Jpeg;

type
  TAPP1Header = packed record
    Marker: Word;                 // APP1 }[J
    Size: Word;                   // TCY
    Magic: array [0.. 5] of Char; // }WbN
    TIFFHeader: DWord;            // TIFF }WbN
  end;

  TImageFileDirectory = packed record
    Tag: Word;
    Format: Word;
    Count: DWord;
    Data: DWord;
  end;

const
  CExifMagic: array [0.. 5] of Char = ('E', 'x', 'i', 'f', #0, #0);

  CTagSOI = $d8ff;
  CTagAPP1 = $e1ff;

  CTIFFMagicM = $2a004d4d;
  CTIFFMagicI = $2a004949;

  CFormatUnsignedByte = 1;
  CFormatString = 2;
  CFormatUnsignedShort = 3;
  CFormatUnsignedLong = 4;
  CFormatUnsignedRational = 5;
  CFormatSignedByte = 6;
  CFormatUndefined = 7;
  CFormatSignedShort = 8;
  CFormatSignedLong = 9;
  CFormatSignedRational = 10;
  CFormatSingleFloat = 11;
  CFormatDoubleFloat = 12;

  CTagImageDescription = $010e;
  CTagMaker = $010f;
  CTagModel = $0110;
  CTagSoftware = $0131;
  CTagDateTime = $0132;
  CTagCopyright = $8298;
  CTagExifIFDPointer = $8769;

  CTagExifWidth = $a002;
  CTagExifHeight = $a003;

{$R *.dfm}

procedure TfrmMain.btnOpenClick(Sender: TObject);
begin
  if (dlgOpen.Execute) then begin
    // C[W̕\
    imgPreview.Picture.LoadFromFile(dlgOpen.FileName);

    // APP1 ̓ǂݍ
    ReadExif(dlgOpen.FileName);
  end;
end;

procedure TfrmMain.ReadExif(const vFileName: String);
var
  APP1Header: TAPP1Header;
  IFD: TImageFileDirectory;
  SOI: Word;
  FS: TFileStream;
  IsBig: Boolean;
  i: Integer;
  TIFFHead: Cardinal;
  SubIFDHead: Cardinal;

  function SwapDWord(const vValue: DWord): DWord;
  begin
    Result :=
      (vValue and $000000ff shl 24) +
      (vValue and $0000ff00 shl 8) +
      (vValue and $00ff0000 shr 8) +
      (vValue and $ff000000 shr 24);
  end;

  function ReadWord: Word;
  begin
    FS.Read(Result, SizeOf(Result));

    if (IsBig) then
      Result := Swap(Result);
  end;

  function ReadDWord: DWord;
  begin
    FS.Read(Result, SizeOf(Result));

    if (IsBig) then
      Result := SwapDWord(Result);
  end;

  procedure ReadIFD;
  begin
    FS.Read(IFD, SizeOf(IFD));

    if (IsBig) then
      with IFD do begin
        Tag := Swap(Tag);
        Format := Swap(Format);
        Count := SwapDWord(Count);
        Data := SwapDWord(Data);
      end;
  end;

  procedure AddStr(const vMsg: String);
  var
    Data: String;
    CurPos: Int64;
  begin
    if (IFD.Format <> CFormatString) then
      Exit;

    SetLength(Data, IFD.Count);

    if (IFD.Count < 5) then
      Move(IFD.Data, PChar(Data)^, Length(Data))
    else begin
      CurPos := FS.Position;
      FS.Seek(TIFFHead + IFD.Data, soBeginning);
      FS.Read(PChar(Data)^, Length(Data));
      FS.Seek(CurPos, soBeginning);
    end;

    memoExif.Lines.Add(vMsg + ': ' + Data);
  end;

  procedure AddUnsignedInt(const vMsg: String);
  begin
    if
      (IFD.Format <> CFormatUnsignedLong) and
      (IFD.Format <> CFormatUnsignedShort)
    then
      Exit;

    memoExif.Lines.Add(vMsg + ': ' + IntToStr(IFD.Data));
  end;

begin
  memoExif.Lines.Clear;

  // APP1 ̓ǂݍ
  FS := TFileStream.Create(vFileName, fmOpenRead);
  try
    FS.Read(SOI, SizeOf(SOI));

    // JPEG ǂ
    if (SOI <> CTagSOI) then
      Exit;

    FS.Read(APP1Header, SizeOf(APP1Header));

    with APP1Header do begin
      // Exif ǂ
      if (Marker <> CTagAPP1) or (Magic <> CExifMagic) then
        Exit;

      // Exif Size 𐳂
      Size := Swap(Size);

      // TIFF Header ărbOg𔻒f
      IsBig := (TIFFHeader = CTIFFMagicM);

      TIFFHead := FS.Position - SizeOf(TIFFHeader);

      // f[^̐擪ɂ킹
      FS.Seek(TIFFHead + ReadDWord, soBeginning);
    end;

    // IFD ǂ
    SubIFDHead := 0;

    for i := 0 to ReadWord - 1 do begin
      ReadIFD;

      with IFD do
        case Tag of
          CTagImageDescription:
            AddStr('摜̐');

          CTagMaker:
            AddStr('[J[');

          CTagModel:
            AddStr('f');

          CTagSoftware:
            AddStr('\tgEFA');

          CTagDateTime:
            AddStr('Be');

          CTagCopyright:
            AddStr('쌠');

          CTagExifIFDPointer:
            if (IFD.Format = CFormatUnsignedLong) then
              SubIFDHead := IFD.Data;
        end;
    end;

    if (SubIFDHead = 0) then
      Exit;

    // Sub IFD ǂ
    FS.Seek(TIFFHead + SubIFDHead, soBeginning);

    for i := 0 to ReadWord - 1 do begin
      ReadIFD;

      with IFD do
        case Tag of
          CTagExifWidth:
            AddUnsignedInt('');

          CTagExifHeight:
            AddUnsignedInt('');
        end;
    end;
  finally
    FS.Free;
  end;
end;

procedure TfrmMain.btnExitClick(Sender: TObject);
begin
  Close;
end;

end.
